@@ -233,7 +233,7 @@ io_frame_leave(struct io *io)
 }
 
 void
-_io_init()
+_io_init(void)
 {
 	static int init = 0;
 
@@ -567,7 +567,8 @@ io_reset(struct io *io, short events, void (*dispatch)(int, short, void*))
 	 */
 	io->flags |= IO_RESET;
 
-	if (event_initialized(&io->ev))
+	if (event_initialized(&io->ev) &&
+	    event_pending(&io->ev, EV_READ|EV_WRITE, NULL))
 		event_del(&io->ev);
 
 	/*
